#!/usr/bin/env python3

# qlub.py
# Python script to manage the QLAN-USB Bridge (QLUB) designed by Martyn Hill
# Derived from the SBASIC program SendFileMQv14_bas, SendFileMQv22_bas

import getopt
import os
import platform
import serial as pyserial
import sys
import time
from typing import BinaryIO, Dict, List, Optional, Tuple


def log(msg: str) -> None:
    sys.stderr.write("qlub: {0}\n".format(msg))


debug_flag: bool = True


def debug(msg: str) -> None:
    if debug_flag:
        log(msg)


def b1(i: int) -> bytes:
    return bytes([i])


def b2(i: int) -> bytes:
    return i.to_bytes(2, byteorder="big")


def b2le(i: int) -> bytes:
    return i.to_bytes(2, byteorder="little")


def b4(i: int) -> bytes:
    return i.to_bytes(4, byteorder="big")


def i2(b: bytes) -> int:
    return int(b[0]) << 8 | b[1]


# MQ constants
mq_signature: bytes = b"%M$Q"
mq_error_mask: int = 32768
mq_long_mask: int = 16384
mq_reply_mask: int = 8192
mq_mid_mask: int = 65536 + ~(mq_error_mask | mq_long_mask | mq_reply_mask)

# net constants
net_peer_types: Dict[str, int] = {"ah": 0, "ql": 0, "zx": 1}

net_profiles: Dict[str, bytes] = {
    "ah": b"\x00\xC8\x4E\x20\x00\x32\x01\xDB\x0B\xB8\x00\x06\x01\xF4\x13\x88\x00\xC8\x00\x00\x00\x8C\x00\xB8\x9C\x40\x05\x00\x01\x0E\x00\xB4",
    "ql": b"\x00\xC8\x4E\x20\x00\x32\x01\xDB\x0B\xB8\x00\x06\x01\xF4\x13\x88\x00\xC8\x00\x00\x00\x8C\x00\xB4\x9C\x40\x05\x00\x01\x0E\x00\xB4",
    "zx": b"\x00\xC8\x4E\x20\x00\x32\x01\xDB\x0B\xB8\x00\x06\x01\xF4\x13\x88\x00\xC8\x00\x00\x00\x8C\x00\xC0\x9C\x40\x05\x00\x01\x27\x00\xB4",
}
net_blk_len: int = 255

# file constants
file_header_lengths: Dict[str, int] = {"ah": 15, "ql": 15, "zx": 9}


def checksum(p: bytes, mod_flag: bool) -> int:
    """calculate a checksum"""

    chk: int = 0
    for i in p:
        chk += i
        if mod_flag:
            chk %= 256
    return chk


def mq_header(
    mid: int, long_flag: bool, dev: int, act: int, par0: int, err: int, len: int
) -> bytes:
    """construct a MQ header"""

    # id
    bmid: bytes = b2(mid)
    if long_flag:
        bmid = bytes([bmid[0] + 64, bmid[1]])

    # header
    hdr: bytes = mq_signature + bmid + b1(dev) + b1(act) + b2(par0) + b2(err) + b2(len)

    # return with checksum
    chk: int = checksum(hdr[0:14], False)
    return hdr + b2(chk)


class Peer:
    """a remote peer (QL, QL with AH ROM or Spectrum)"""

    def __init__(self, type: str, stn: int):
        self.type: str = type
        self.stn: int = stn
        self.pty: int = net_peer_types[type]
        self.profile: bytes = net_profiles[type]
        self.header_length: int = file_header_lengths[type]

    def __repr__(self) -> str:
        return 'Peer("{0}", {1})'.format(self.type, self.stn)


class Bridge:
    def __init__(self, port: str, stn: int):

        self.port: str = port
        self.stn: int = stn
        self.next_mid: int = 1

    def connect(self) -> None:
        """connect via serial"""

        # open it
        try:
            self.ser: BinaryIO = pyserial.Serial(
                port=self.port, baudrate=115200, rtscts=True, timeout=100
            )
        except pyserial.SerialException as err:
            log("connect: {0}".format(err))
            sys.exit(1)

    def get_next_mid(self) -> int:
        """generate the next message ID"""

        mid: int = self.next_mid
        self.next_mid += 1
        if self.next_mid > 8191:
            self.next_mid = 1
        return mid

    def mq_read(self, mid: int, timeout: int) -> Tuple[int, Optional[bytes]]:
        """read a message"""

        payload: Optional[bytes] = None

        hdr: bytearray = bytearray(16)
        return_code: int = 0

        while True:

            # read signature
            start: float = time.time()
            idx: int = 0
            while True:

                # check for timeout
                end: float = time.time()
                if (end - start) > timeout and idx == 0:
                    return (-1, None)

                # read byte if present
                ch: bytes = self.ser.read(1)
                if ch == b"":
                    continue

                # validate this byte is part of signature, reset if not
                if ch[0] != mq_signature[idx]:
                    idx = 0
                    continue

                # add signature char into header
                hdr[idx] = ch[0]
                idx += 1
                if idx >= len(mq_signature):
                    break

            # read rest of header
            for idx in range(4, 16):

                while True:

                    # check for timeout
                    end = time.time()
                    if (end - start) > timeout and idx == 4:
                        return_code = -2
                        break

                    # read char
                    ch = self.ser.read(1)
                    if ch != b"":
                        break

                hdr[idx] = ch[0]

            # validate checksum
            chk: int = checksum(hdr[0:14], False)
            reply_chk: int = i2(hdr[14:16])
            if reply_chk != chk:
                return_code = -3
                break

            # derive flags from mid field
            reply_mid: int = i2(hdr[4:6])
            error_flag: bool = (reply_mid & mq_error_mask) != 0
            long_flag: bool = (reply_mid & mq_long_mask) != 0
            reply_flag: bool = (reply_mid & mq_reply_mask) != 0
            reply_mid &= mq_mid_mask

            # wrong mid
            if reply_mid != mid:
                return_code = -4
                break

            # error
            if error_flag:
                err: int = i2(hdr[10:12])
                debug("mq_read mid={0} err={1}".format(mid, err))
                return_code = -5
                break

            # not a reply
            if not reply_flag:
                return_code = -6
                break

            # message
            if long_flag:
                payload = self.ser.read(i2(hdr[12:14]))
                return_code = 1

            # ack
            else:
                return_code = 0

            break

        debug("mq_read mid={0} ret={1}".format(mid, return_code))
        return (return_code, payload)

    def mq_write(
        self, dev: int, act: int, p0: int, err: int, p2: int, payload: bytes
    ) -> int:
        """write a message"""

        mid: int = self.get_next_mid()
        hdr: bytes = mq_header(mid, True, dev, act, p0, err, p2)
        self.ser.write(hdr + payload)
        self.ser.flush()
        return mid

    def mq_send(self, dev: int, act: int, p0: int, err: int, payload: bytes) -> int:
        """send a block"""

        send_pending = True
        while True:

            # send block
            if send_pending:
                mid = self.mq_write(dev, act, p0, err, len(payload), payload)
                send_pending = False

            # wait for ack
            (result, p) = self.mq_read(mid, 500)
            if result == 0:
                break

            # retry send
            if result == -3 or result == -5:
                send_pending = True

        debug("mq_send ret={0}".format(result))
        return result

    def mq_recv(self, src: int, dst: int, blk: int) -> bytes:
        """receive a block"""

        # solicit a block
        mid: int = self.mq_write(2, 2, 0, 256 * dst + src, blk, b"")

        # read the block back
        while True:

            (result, payload) = self.mq_read(mid, 20)

            # error
            if result == -1:
                continue

            # error or short
            if payload is None:
                debug("mq_recv: mid={0} res={1}".format(mid, result))
                continue

            if len(payload) < 9:
                debug(
                    "mq_recv: mid={0} res={1} len={2}".format(mid, result, len(payload))
                )
                continue

            # return block
            return payload

    def mq_set_profile(self, peer: Peer) -> int:
        """set the timing profile"""

        result: int = self.mq_send(2, 4, 0, peer.pty, peer.profile)
        if result != 0:
            return result

        result = self.mq_send(2, 5, 0, peer.pty, b"")
        return result

    def mq_status(self) -> str:
        """fetch status"""

        mid: int = bridge.mq_write(0, 1, 0, 0, 0, b"")
        (result, payload) = bridge.mq_read(mid, 20)
        if result > 0 and payload is not None:
            return payload.decode("utf-8")
        return "mq_status: res={0}".format(result)

    def disconnect(self) -> None:
        """disconnect serial"""

        self.ser.close()


def file_header_ql(len: int, typ: int, dsp: int, ext: int) -> bytes:
    """15-byte QL file header"""

    return b"\xff" + b4(len) + b"\x00" + b1(typ) + b4(dsp) + b4(ext)


def file_header_zx(typ: int, len: int, sta: int, pln: int, lno: int) -> bytes:
    """9-byte Spectrum file header"""

    return b1(typ) + b2le(len) + b2le(sta) + b2le(pln) + b2le(lno)


def net_header(dst: int, src: int, blk: int, typ: int, len: int, chk: int) -> bytes:
    """8-byte net block header"""

    # header
    hdr: bytes = b1(dst) + b1(src) + b2le(blk) + b1(typ) + b1(len) + b1(chk)

    # return with checksum
    hchk: int = checksum(hdr, True)
    return hdr + b1(hchk)


def net_block(block: bytes, bno: int, src: int, dst: int, typ: int) -> bytes:
    """prepend a block header to a block"""

    chk: int = checksum(block, True)
    nhdr: bytes = net_header(dst, src, bno, typ, len(block), chk)
    return nhdr + block


def net_payload(
    fil: BinaryIO, bno: int, blen: int, hdr: bytes, src: int, dst: int, typ: int
) -> bytes:
    """read a block (including file header) and prepend a block header"""

    # read from file less file header
    blk: bytes
    if bno == 0:
        blk = hdr + fil.read(blen - len(hdr))
    else:
        blk = fil.read(blen)

    # return with net header
    return net_block(blk, bno, src, dst, typ)


def send_file(bridge: Bridge, peer: Peer, fnm: str, file_hdr: bytes) -> None:
    """send a file, prepending a file header"""

    # count blocks and get last block length
    file_len: int = os.path.getsize(fnm)
    file_hdr_len: int = len(file_hdr)
    blk_cnt: int = int((file_hdr_len + file_len) / net_blk_len)
    last_blk_len: int = (file_hdr_len + file_len) - (blk_cnt * net_blk_len)
    if last_blk_len > 0:
        last_blk = blk_cnt
    else:
        last_blk = blk_cnt - 1
        last_blk_len = net_blk_len

    # open file
    file: BinaryIO = open(fnm, "rb")

    # net profile
    result: int = bridge.mq_set_profile(peer)
    if result != 0:
        debug("send_file: mq_set_profile returned {0}".format(result))

    # send all blocks
    for blk in range(last_blk + 1):

        # last block?
        if blk == last_blk:
            blk_len = last_blk_len
            net_type = 1
        else:
            blk_len = net_blk_len
            net_type = 0

        # send
        payload: bytes = net_payload(
            file, blk, blk_len, file_hdr, bridge.stn, peer.stn, net_type
        )
        bridge.mq_send(2, 3, 0, 0, payload)

    # close file
    file.close()


def send_file_ql(
    bridge: Bridge, peer: Peer, name: str, typ: int, dsp: int, ext: int
) -> None:
    """send a file to a QL peer, prepending header with specified params"""

    hdr: bytes = file_header_ql(os.path.getsize(name), typ, dsp, ext)
    send_file(bridge, peer, name, hdr)


def send_file_zx(
    bridge: Bridge, peer: Peer, name: str, typ: int, sta: int, pln: int, lno: int
) -> None:
    """send a file to a Spectrum peer, prepending header with specified params"""

    hdr: bytes = file_header_zx(typ, os.path.getsize(name), sta, pln, lno)
    send_file(bridge, peer, name, hdr)


def send_file_zx_screen(bridge: Bridge, peer: Peer, name: str) -> None:
    """send a SCREEN$ to a Spectrum peer, prepending header"""

    send_file_zx(bridge, peer, name, 3, 16384, 0, 0)


def send_data(bridge: Bridge, peer: Peer, data: bytes) -> None:
    """send a file, prepending a file header"""

    # count blocks and get last block length
    file_len: int = len(data)
    blk_cnt: int = int(file_len / net_blk_len)
    last_blk_len: int = file_len - (blk_cnt * net_blk_len)
    if last_blk_len > 0:
        last_blk = blk_cnt
    else:
        last_blk = blk_cnt - 1
        last_blk_len = net_blk_len

    idx: int = 0

    # net profile
    result: int = bridge.mq_set_profile(peer)
    if result != 0:
        debug("send_file: mq_set_profile returned {0}".format(result))

    # send all blocks
    for blk in range(last_blk + 1):

        # last block?
        if blk == last_blk:
            blk_len = last_blk_len
            net_type = 1
        else:
            blk_len = net_blk_len
            net_type = 0

        # send
        block = data[idx : idx + blk_len]
        payload: bytes = net_block(block, blk, bridge.stn, peer.stn, net_type)

        bridge.mq_send(2, 3, 0, 0, payload)
        idx += blk_len


def read_net_hdr(payload: bytes) -> Tuple[int, bytes]:
    """split net payload into header and body"""

    # read the net header
    net_hdr: bytes = payload[0:8]
    nd_type: int = net_hdr[4]
    net_blk: bytes = payload[8:]
    return (nd_type, net_blk)


def recv_file(bridge: Bridge, peer: Peer, name: str, file_hdr_len: int) -> None:
    """receive a file, strip a header of specified length"""

    file: BinaryIO = open(name, "wb")

    bridge.mq_set_profile(peer)

    blk: int = 0
    while True:

        # receive a block
        payload: bytes = bridge.mq_recv(peer.stn, bridge.stn, blk)

        # strip the file header and write
        (nd_type, net_blk) = read_net_hdr(payload)
        file_hdr_flag: bool = blk == 0
        file.write(net_blk[file_hdr_len * file_hdr_flag :])

        # last block?
        if nd_type:
            break

        blk += 1

    file.flush()
    file.close()


def recv_data(bridge: Bridge, peer: Peer) -> bytes:
    """receive data, return as a bytearray"""

    bridge.mq_set_profile(peer)

    data: bytes = b""
    blk: int = 0
    while True:

        # receive a block
        payload: bytes = bridge.mq_recv(peer.stn, bridge.stn, blk)

        # strip the file header and append the block
        (nd_type, net_blk) = read_net_hdr(payload)
        data += net_blk

        # last block?
        if nd_type:
            break

        blk += 1

    return data


def status(bridge: Bridge) -> None:
    """fetch status and print"""

    print(bridge.mq_status())


def gopher_line(type: str, label: str, selector: str, host: str, port: int) -> bytes:
    """format a gopher selector line"""

    return "{0}{1}\t{2}\t{3}\t{4}\r\n".format(type, label, selector, host, port).encode(
        "utf-8"
    )


def gopher_dir(bridge: Bridge, peer: Peer, dir: str) -> None:
    """write a gophermap for the dir"""

    map: bytes = b""
    for file in os.listdir(".{0}".format(dir)):
        selector: str = "{0}/{1}".format(dir, file)
        if os.path.isdir(".{0}".format(selector)):
            map += gopher_line("1", file, selector, str(bridge.stn), 70)
        elif os.path.isfile(".{0}".format(selector)):
            map += gopher_line("9", file, selector, str(bridge.stn), 70)

    map += b".\r\n"
    send_data(bridge, peer, map)


def gopher_error(bridge: Bridge, peer: Peer, msg: str) -> None:

    send_data(bridge, peer, gopher_line("3", msg, "fake", "(NULL)", 0) + b".\r\n")


def gopher(bridge: Bridge, peer: Peer) -> None:
    """basic gopher server"""

    while True:

        # read input
        line: bytes = recv_data(bridge, peer)
        crlf: int = line.find(b"\r\n")

        # home dir
        if crlf == 0:
            gopher_dir(bridge, peer, "")
            continue

        # validate
        if crlf < 0 or line[0:1] != b"/" or line.find(b"..") != -1:
            gopher_error(bridge, peer, "Invalid selector")
            continue

        # file or dir
        name: str = line[:crlf].decode("utf-8")
        debug("gopher name={0}".format(name))
        file = ".{0}".format(name)
        if os.path.isdir(file):
            gopher_dir(bridge, peer, name)
        elif os.path.isfile(file):
            send_file(bridge, peer, file, b"")
        else:
            gopher_error(bridge, peer, "File not found")


def default_port_name() -> str:
    """derive expected serial port name from OS"""

    system: str = platform.system()
    if system[0:6] == "CYGWIN":
        return "/dev/ttyS2"
    elif system == "Darwin":
        return "/dev/tty.usbmodem123451"
    elif system == "Linux":
        return "/dev/ttyACM0"
    elif system == "Windows":
        return "COM3"
    else:
        log("port name not known for system {0}".format(system))
        sys.exit(1)


def write_bytes(b: bytes) -> None:
    """utility method to output byte array"""

    sys.stdout.buffer.write(b)
    sys.stdout.buffer.flush()


def valid_peer(arg: str) -> bool:

    return arg == "ah" or arg == "ql" or arg == "zx"


def valid_ql_peer(arg: str) -> bool:

    return arg == "ah" or arg == "ql"


def usage() -> None:

    sys.stderr.write("Usage: %s [-p <prt> -s <stn>] gopher ah|ql|zx <stn>\n" % argv[0])
    sys.stderr.write(
        "       %s                     header ah|ql    <len> <typ> <dsp> <ext>\n"
        % argv[0]
    )
    sys.stderr.write(
        "       %s                     header zx       <typ> <len> <sta> <prl> <lno>\n"
        % argv[0]
    )
    sys.stderr.write(
        "       %s [-p <prt> -s <stn>] recv   ah|ql|zx <src> <name> [header]\n"
        % argv[0]
    )
    sys.stderr.write(
        "       %s [-p <prt> -s <stn>] send   ah|ql    <dst> <name> [header <typ> <dsp> <ext>]\n"
        % argv[0]
    )
    sys.stderr.write(
        "       %s [-p <prt> -s <stn>] send   zx       <dst> <name> [header <typ> <sta> <prl> <lno>]\n"
        % argv[0]
    )
    sys.stderr.write("       %s [-p <prt>]          status\n" % argv[0])
    sys.exit(1)


# defaults
port: str = default_port_name()
station: int = 1

# begin parsing command line arguments
argv: List[str] = sys.argv
argc: int = len(argv)

# options -p, -s
try:
    opts, args = getopt.gnu_getopt(argv, "p:s:", [])
except getopt.GetoptError as err:
    print(err)
    usage()
    sys.exit(1)

for o, a in opts:
    if o == "-p":
        port = a
    elif o == "-s":
        station = int(a)
    else:
        assert False, "unhandled option"

argv = args
argc = len(argv)

# gopher
if argc == 4 and argv[1] == "gopher" and valid_peer(argv[2]):
    bridge = Bridge(port, station)
    bridge.connect()
    gopher(bridge, Peer(argv[2], int(argv[3])))
    bridge.disconnect()

# header utility
elif argc == 7 and argv[1] == "header" and valid_ql_peer(argv[2]):
    write_bytes(file_header_ql(int(argv[3]), int(argv[4]), int(argv[5]), int(argv[6])))

elif argc == 8 and argv[1] == "header" and argv[2] == "zx":
    write_bytes(
        file_header_zx(
            int(argv[3]), int(argv[4]), int(argv[5]), int(argv[6]), int(argv[7])
        )
    )

# recv
elif argc == 5 and argv[1] == "recv" and valid_peer(argv[2]):
    bridge = Bridge(port, station)
    bridge.connect()
    recv_file(bridge, Peer(argv[2], int(argv[3])), argv[4], 0)
    bridge.disconnect()

elif argc == 6 and argv[1] == "recv" and valid_peer(argv[2]) and argv[5] == "header":
    bridge = Bridge(port, station)
    bridge.connect()
    peer: Peer = Peer(argv[2], int(argv[3]))
    recv_file(bridge, peer, argv[4], peer.header_length)
    bridge.disconnect()

# send
elif argc == 5 and argv[1] == "send" and valid_peer(argv[2]):
    bridge = Bridge(port, station)
    bridge.connect()
    send_file(bridge, Peer(argv[2], int(argv[3])), argv[4], b"")
    bridge.disconnect()

elif argc == 9 and argv[1] == "send" and valid_ql_peer(argv[2]) and argv[5] == "header":
    bridge = Bridge(port, station)
    bridge.connect()
    send_file_ql(
        bridge,
        Peer(argv[2], int(argv[3])),
        argv[4],
        int(argv[6]),
        int(argv[7]),
        int(argv[8]),
    )
    bridge.disconnect()

elif argc == 10 and argv[1] == "send" and argv[2] == "zx" and argv[5] == "header":
    bridge = Bridge(port, station)
    bridge.connect()
    send_file_zx(
        bridge,
        Peer(argv[2], int(argv[3])),
        argv[4],
        int(argv[6]),
        int(argv[7]),
        int(argv[8]),
        int(argv[9]),
    )
    bridge.disconnect()

# status
elif argc == 2 and argv[1] == "status":
    bridge = Bridge(port, station)
    bridge.connect()
    status(bridge)
    bridge.disconnect()

elif argc == 4 and argv[1] == "test" and valid_ql_peer(argv[2]):
    bridge = Bridge(port, station)
    bridge.connect()
    send_data(bridge, Peer(argv[2], int(argv[3])), b"Hello World\r\n")
    bridge.disconnect()

else:
    usage()
